/*
This is a part of the LiteStep Shell Source code.

Copyright (C) 1997-2001 The LiteStep Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/****************************************************************************
- Fixed [systray2 / Message / 2001-11-26]
  - Stripped everything out of it except for basic gui, this modulewill not run
    without a new build

- Fixed [systray2 / Message / 2001-11-21]
  - Added Windows2000 and WindowxXP support. Seems to work fine when first loaded,
    but after a recycle w2k loses its volume icon and xp loses all icons (?). A simple fix
    for w2k is to run trayfix.exe after a recycle, and that will put the volume icon back.
    
    NOTES:
    The DUN icon uses 4 icons to display status, when NIM_ADD the icons, they all have the
    same HWND, but differ in NOTIFYICONDATA.uID. To get around this, we only check to see
    if an icon already exists by HWND only, and not uID. On NIM_MODIFY, we set the uID
    of any found icon to the modifying icon's uID or problems occur with the mouse handling.
    
    UnloadShellServiceObjects needs to be called before LM_SAVESYSTRAY is handled or
    the DUN/Volume icons get saved also and won't be handled properly when the tray is
    restarted after a recycle.
    
    The tooltip on the DUN icon still doesn't handle well. I think this is our implementation
    and not MS related.
****************************************************************************/
#include "systray.h"

// defined directions
#define TD_VERTICAL    0x1
#define TD_TOTOPLEFT   0x2

#define TD_RIGHT       0x0
#define TD_LEFT        (TD_TOTOPLEFT)
#define TD_UP          (TD_TOTOPLEFT|TD_VERTICAL)
#define TD_DOWN        (TD_VERTICAL)

// name of shelldesktop class
#define WC_SHELLDESKTOP    "DesktopBackgroundClass"

const char szAppName[] = "systray"; // Our window class, etc

const char rcsRevision[] = "$Revision: 1.5 $"; // Our Version*/
const char rcsId[] = "$Id: systray.cpp,v 1.5 2001/11/26 22:59:44 message Exp $"; // The Full RCS ID.*/

systray *tray; // The module

//=========================================================
// Initialization and cleanup
//=========================================================

int initModuleEx(HWND parentWnd, HINSTANCE dllInst, LPCSTR szPath)
{
	int code;
	Window::init(dllInst);
	tray = new systray(parentWnd, code);
	return code;
}
int initWharfModule(HWND parentWnd, HINSTANCE dllInst, void *wharfData)
{
	int code;
	Window::init(dllInst);
	tray = new systray(parentWnd, code);
	return code;
}
void quitModule(HINSTANCE dllInst)
{
	delete tray;
}
void quitWharfModule(HINSTANCE dllInst)
{
	delete tray;
}

// BANG COMMAND FUNCTIONS

void bangShow( HWND sender, const char *args )
{
	tray->show();
}
void bangHide( HWND sender, const char *args )
{
	tray->hide();
}
void bangToggle( HWND sender, const char *args )
{
	tray->toggle();
}
void bangOnTop( HWND sender, const char *args )
{
	tray->toggleOnTop();
}
void bangMove( HWND sender, const char *args )
{
	char szX[MAX_LINE_LENGTH], szY[MAX_LINE_LENGTH];
	char *tokens[2] = {szX, szY};
	LCTokenize( args, tokens, 2, NULL );
	tray->move(atoi(szX), atoi(szY));
}
void bangSize( HWND sender, const char *args )
{
	char szCX[MAX_LINE_LENGTH], szCY[MAX_LINE_LENGTH];
	char *tokens[2] = {szCX, szCY};
	LCTokenize( args, tokens, 2, NULL );
	tray->size(atoi(szCX), atoi(szCY));
}

//=========================================================
// Module code
//=========================================================

//---------------------------------------------------------
// Module constructor
//---------------------------------------------------------
systray::systray(HWND parentWnd, int& code):
		Window(szAppName)
{
	// reading settings for window creation
	// ======================================

	// get ls window handle
	liteStep = GetLitestepWnd();
	desktop = FindWindow(WC_SHELLDESKTOP, NULL);
	if (!desktop)
		desktop = GetDesktopWindow();

	// if the parent window is not litestep, it must be the wharf
	inWharf = (parentWnd != liteStep);

	// get the position and size of the _tmain window
	// using different method if it's docked in the wharf
	if ( inWharf )
	{
		RECT rc;
		int wharfBorder;

		// get the wharf window size
		GetClientRect( parentWnd, &rc );
		wharfBorder = GetRCInt( "WharfBevelWidth", 0 );

		trayX = wharfBorder;
		trayY = wharfBorder;
		trayWidth = rc.right - 2 * wharfBorder;
		trayHeight = rc.bottom - 2 * wharfBorder;
	}
	else
	{
		trayX = GetRCInt( "systrayX", 0 );
		if ( trayX < 0 )
			trayX += GetSystemMetrics(SM_CXSCREEN);
		trayY = GetRCInt( "systrayY", 0 );
		if ( trayY < 0 )
			trayY += GetSystemMetrics(SM_CYSCREEN);
		trayWidth = GetRCInt( "systrayWidth", 64 );
		trayHeight = GetRCInt( "systrayHeight", 64 );
	}

	if (!createWindow(WS_EX_TOOLWINDOW, szAppName, inWharf ? WS_CHILD : WS_POPUP,
	                  trayX, trayY, trayWidth, trayHeight, (inWharf ? parentWnd : desktop)))
	{

		RESOURCE_MSGBOX(hInstance, IDS_SYSTRAY2_ERROR3,
		                "Unable to create window.", szAppName)

		code = 1;
		return ;
	}

	PostMessage(liteStep, LM_SYSTRAYREADY, NULL, NULL);

	code = 0;
}

systray::~systray()
{
	destroyWindow();
}

void systray::windowProc(Message& message)
{
	BEGIN_MESSAGEPROC
	MESSAGE(onCreate, WM_CREATE)
	MESSAGE(onDestroy, WM_DESTROY)
	MESSAGE(onEndSession, WM_ENDSESSION)
	MESSAGE(onEndSession, WM_QUERYENDSESSION)
	MESSAGE(onGetRevId, LM_GETREVID)
	MESSAGE(onSysTray, LM_SYSTRAY)
	MESSAGE(onWindowPosChanging, WM_WINDOWPOSCHANGING)
	MESSAGE(onKeyMessage, WM_KEYDOWN)
	MESSAGE(onKeyMessage, WM_KEYUP)
	MESSAGE(onKeyMessage, WM_HOTKEY)
	MESSAGE(onMouse, WM_LBUTTONDOWN)
	MESSAGE(onMouse, WM_MBUTTONDOWN)
	MESSAGE(onMouse, WM_RBUTTONDOWN)
	MESSAGE(onMouse, WM_LBUTTONUP)
	MESSAGE(onMouse, WM_MBUTTONUP)
	MESSAGE(onMouse, WM_RBUTTONUP)
	MESSAGE(onMouse, WM_LBUTTONDBLCLK)
	MESSAGE(onMouse, WM_MBUTTONDBLCLK)
	MESSAGE(onMouse, WM_RBUTTONDBLCLK)
	MESSAGE(onMouse, WM_MOUSEMOVE)
	MESSAGE(onPaint, WM_PAINT)
	MESSAGE(onPaint, LM_REPAINT)
	MESSAGE(onSysCommand, WM_SYSCOMMAND)
	END_MESSAGEPROC
}

void systray::onEndSession(Message& message)
{
	message.lResult = SendMessage(liteStep, message.uMsg, message.wParam, message.lParam);
}
void systray::onKeyMessage(Message& message)
{
	// Forward these messages
	PostMessage(liteStep, message.uMsg, message.wParam, message.lParam);
}
void systray::onSysCommand(Message& message)
{
	if (message.wParam == SC_CLOSE)
		PostMessage(liteStep, WM_KEYDOWN, LM_SHUTDOWN, 0);
	else
		message.lResult = DefWindowProc(hWnd, message.uMsg, message.wParam, message.lParam);
}

void systray::onCreate(Message& message)
{
	char szTemp[256];
	int tmpInteger = 0;
	int msgs[] = {LM_GETREVID, LM_SYSTRAY, 0};

	// register messages
	SendMessage(liteStep, LM_REGISTERMESSAGE, (WPARAM)hWnd, (LPARAM)msgs);

	// has to be done
	SetWindowLong(hWnd, GWL_USERDATA, magicDWord);

	// make sure it handle dblclicks... =)
	SetClassLong(hWnd, GCL_STYLE, CS_DBLCLKS | GetClassLong(hWnd, GCL_STYLE));

	// load settings here
	// ===========================================
	visible = inWharf || GetRCBool("systrayHidden", FALSE);
	onTop = !inWharf && GetRCBool("systrayAlwaysOnTop", TRUE);
	hideIfEmpty = !inWharf && GetRCBool("systrayHideIfEmpty", TRUE);
	autoSize = !inWharf && GetRCBool("systrayAutoSize", TRUE);
	tmpInteger = GetRCInt("systrayBorderSize", 0);
	borderTop = max(GetRCInt("systrayBorderTop", tmpInteger), 0);
	borderLeft = max(GetRCInt("systrayBorderLeft", tmpInteger), 0);
	borderRight = max(GetRCInt("systrayBorderRight", tmpInteger), 0);
	borderBottom = max(GetRCInt("systrayBorderBottom", tmpInteger), 0);
	borderDrag = !inWharf && GetRCBool("systrayBorderDrag", TRUE);
	snapDistance = max(GetRCInt("systraySnapDistance", (GetRCBool("systraySnap", TRUE) ? 10 : 0)), 0);

	// colors
	clrBack = GetRCColor("systrayBGColor", RGB(255, 255, 255));
	clrBorder = GetRCColor("systrayBorderColor", RGB(0, 0, 0));
	clrBorderDk = GetRCColor("sysTrayBorderDark", clrBorder);

	// OnAdd and OnDel commands
	GetRCString("systrayOnAdd", szTemp, "", 256);
	if ( szTemp[0] != '\0' ) {
		OnAddCommand = new char[strlen(szTemp) + 1];
		strcpy(OnAddCommand, szTemp);
	}
	else {
		OnAddCommand = NULL;
	}

	GetRCString("sysTrayOnDel", szTemp, "", 256);
	if ( szTemp[0] != '\0' ) {
		OnDelCommand = new char[strlen(szTemp) + 1];
		strcpy(OnDelCommand, szTemp);
	}
	else {
		OnDelCommand = NULL;
	}
	FILE* f;
	f = LCOpen(NULL);

	char token1[256], token2[256], token3[256], extra[256];
	char* tokens[3] = { token1, token2, token3};
	char* tempNew;
	int tmpMark, Mark = 1;

	while ( LCReadNextConfig (f, "*sysTrayOn", szTemp, sizeof (szTemp) ) ) {

		LCTokenize (szTemp, tokens, 3, extra);
		tmpMark = atoi(token2);

		if ( tmpMark == Mark ) {
			tempNew = new char[strlen(token3) + 1];
			strcpy(tempNew, token3);
			OnNumList.push_back(tempNew);
		}
		else {
			for ( ; Mark < tmpMark ; Mark++ ) {
				tempNew = new char[6];
				strcpy(tempNew, "!none");
				OnNumList.push_back(tempNew);
			}
			tempNew = new char[strlen(token3) + 1];
			strcpy(tempNew, token3);
			OnNumList.push_back(tempNew);
		}

		Mark++;
	}

	LCClose(f);

	OnNumCount = 0;
	OnNumExe = 1;

	// skinning
	GetRCString("systrayBitmap", szTemp, "" , 256);

	if ( szTemp[0] == '\0' )
	{
		hbmSkin = NULL;
		transpSkin = (clrBack == RGB(255, 0, 255));
		if (!transpSkin)
			transpSkin = (borderTop + borderLeft + borderRight + borderBottom > 0) && (clrBorder == RGB(255, 0, 255));
	}
	else
	{
		BITMAP bm;
		HRGN hBitmapRgn, hOpaqueRgn;

		hbmSkin = LoadLSImage(szTemp, NULL);

		// get info from the bitmap
		GetObject( hbmSkin, sizeof(BITMAP), &bm );

		hBitmapRgn = BitmapToRegion( hbmSkin, RGB(255, 0, 255), 0x101010, 0, 0);
		hOpaqueRgn = CreateRectRgn( 0, 0, bm.bmWidth, bm.bmHeight );

		// to see if any pixels are transparent
		transpSkin = !EqualRgn( hBitmapRgn, hOpaqueRgn );

		DeleteObject( hBitmapRgn );
		DeleteObject( hOpaqueRgn );
	}
	skinTiled = GetRCBool("systrayBitmapTiled", TRUE);

	// initiate vars for the background creation
	hbmBack = NULL;
	transpBack = FALSE;

	//tray wrapping
	wrapCount = max(GetRCInt("systrayWrapCount", 0), 0);

	// get direction
	GetRCString("systrayDirection", szTemp, "", 256);
	if ( !::strnicmp(szTemp, "left", 5) )
		direction = TD_LEFT;
	else if ( !::strnicmp(szTemp, "up", 3) )
		direction = TD_UP;
	else if ( !::strnicmp(szTemp, "down", 5) )
		direction = TD_DOWN;
	else
		direction = TD_RIGHT;

	// get wrap-direction
	GetRCString("systrayWrapDirection", szTemp, "", 256);
	if (direction & TD_VERTICAL)
	{
		if ( !::strnicmp( szTemp, "left", 5 ) )
			wrapDirection = TD_LEFT;
		else
			wrapDirection = TD_RIGHT;
	}
	else
	{
		if ( !::strnicmp( szTemp, "up", 3) )
			wrapDirection = TD_UP;
		else
			wrapDirection = TD_DOWN;
	}

	// icon
	iconSize = max(GetRCInt( "systrayIconSize", 16 ), 0);
	iconSpacingX = max(GetRCInt( "systrayIconSpacingX", 1 ), 0);
	iconSpacingY = max(GetRCInt( "systrayIconSpacingY", 1 ), 0);

	// this section evaluates how the placing of new icons should be calculated
	deltaX = ((( direction & TD_VERTICAL ) ? wrapDirection : direction ) == TD_LEFT) ? -1 : 1;
	deltaY = ((( direction & TD_VERTICAL ) ? direction : wrapDirection ) == TD_DOWN) ? 1 : -1;

	deltaX *= iconSize + iconSpacingX;
	deltaY *= iconSize + iconSpacingY;

	setFirstLast();
	createBackground();

	// NOT FOR WHARF
	if ( !inWharf )
	{

		AddBangCommand( "!systrayhide", bangHide );
		AddBangCommand( "!systrayshow", bangShow );
		AddBangCommand( "!systraytoggle", bangToggle );
		AddBangCommand( "!systrayontop", bangOnTop );
		AddBangCommand( "!systraymove", bangMove );
		AddBangCommand( "!systraysize", bangSize );

	}

	// set always on top
	if (onTop)
		SetWindowPos( hWnd, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE );

	// show window
	ShowWindow( hWnd, (!visible || hideIfEmpty) ? SW_HIDE : SW_SHOWNOACTIVATE );
	InvalidateRect( hWnd, NULL, TRUE );

	// create a tooltip window
	tooltip = CreateWindowEx(WS_EX_TOOLWINDOW | WS_EX_TOPMOST, TOOLTIPS_CLASS, NULL, TTS_ALWAYSTIP | WS_POPUP, 0, 0, 0, 0, hWnd, NULL, NULL, NULL);

	uLastID = 1;
	pFirst = NULL;
	pLast = NULL;
}

void systray::onDestroy(Message& message)
{
	int msgs[] = {LM_GETREVID, LM_SYSTRAY, 0};

	// NOT FOR WHARF
	if ( !inWharf )
	{
		RemoveBangCommand("!systrayHide");
		RemoveBangCommand("!systrayShow");
		RemoveBangCommand("!systrayToggle");
		RemoveBangCommand("!systrayOnTop");
		RemoveBangCommand("!systrayMove");
		RemoveBangCommand("!systraySize");
	}

	SendMessage(liteStep, LM_UNREGISTERMESSAGE, (WPARAM)hWnd, (LPARAM)msgs);

	// destroy the list of icons
	iconListDelAll();

	// destroy the tooltip window
	if (tooltip)
		DestroyWindow( tooltip );
	// delete gdi objects...
	if (hbmSkin)
		DeleteObject( hbmSkin );
	if (hbmBack)
		DeleteObject( hbmBack );
	if (hrgnBack)
		DeleteObject( hrgnBack );

}

void systray::onGetRevId(Message& message)
{
	LPSTR buf = (LPSTR)(message.lParam);

	switch (message.wParam)
	{
		case 0:
		sprintf(buf, "systray2.dll: %s", &rcsRevision[11]);
		buf[strlen(buf) - 1] = '\0';
		break;
		case 1:
		strcpy(buf, &rcsId[1]);
		buf[strlen(buf) - 1] = '\0';
		break;
		default:
		strcpy(buf, "");
	}
	message.lResult = strlen(buf);
}


void systray::onSysTray(Message& message)
{
	switch ( message.wParam )
	{
		case NIM_ADD:
		{
			message.lResult = (BOOL)( iconAdd((PLSNOTIFYICONDATA)message.lParam ) != NULL);
			break;
		}
		case NIM_MODIFY:
		{
			message.lResult = (BOOL)( iconMod((PLSNOTIFYICONDATA)message.lParam ) != NULL);
			break;
		}
		case NIM_DELETE:
		{
			message.lResult = (BOOL)iconDel((PLSNOTIFYICONDATA)message.lParam);
			break;
		}
		default:
		{
			message.lResult = FALSE;
			break;
		}
	}
}

void systray::onMouse(Message& message)
{
	if (!onMouseIcon(message))
	{
		int x = message.lParamLo, y = message.lParamHi;
		if (message.uMsg == WM_LBUTTONDOWN)
		{
			if ( borderDrag && ( (x <= borderLeft) || (x >= trayWidth - borderRight) ||
			                     (y <= borderTop) || (y >= trayHeight - borderBottom) ) )
				SendMessage( hWnd, WM_SYSCOMMAND, SC_MOVE | 2, message.lParam );
		}
		else if (message.uMsg == WM_RBUTTONUP)
		{
			POINT point = {x, y};

			ClientToScreen(hWnd, &point);
			PostMessage( liteStep, LM_POPUP, 0, MAKELPARAM(point.x, point.y) );
		}
	}
	message.lResult = DefWindowProc(hWnd, message.uMsg, message.wParam, message.lParam);
}

BOOL systray::onMouseIcon(Message& message)
{
	PSYSTRAYICON pSysTrayIcon;
	MSG msg;
	POINT pt;
	pt.x = message.lParamLo;
	pt.y = message.lParamHi;

	// get icon at point
	pSysTrayIcon = iconListFindPt(pt);

	//SetForegroundWindow(pSysTrayIcon->hWnd);

	msg.hwnd = hWnd;
	msg.message = message.uMsg;
	msg.wParam = message.wParam;
	msg.lParam = message.lParam;
	msg.time = GetTickCount();
	msg.pt = pt;
	SendMessage( tooltip, TTM_RELAYEVENT, 0, (LPARAM) &msg );

	// remove "unwanted" icons
	iconListCleanup();

	// post message to that icon
	if (pSysTrayIcon && pSysTrayIcon->uCallbackMessage)
	{
		SendMessage( pSysTrayIcon->hWnd, pSysTrayIcon->uCallbackMessage, (WPARAM)pSysTrayIcon->uID, (LPARAM)message.uMsg );
		return TRUE;
	}
	return FALSE;
}

void systray::onWindowPosChanging(Message& message)
{
	LPWINDOWPOS lpwp = (LPWINDOWPOS)message.lParam;

	if ( !( lpwp->flags & SWP_NOSIZE ) )
	{
		if ( (trayWidth != lpwp->cx) || (trayHeight != lpwp->cy) )
		{
			trayWidth = lpwp->cx;
			trayHeight = lpwp->cy;
		}
	}

	if ( !( lpwp->flags & SWP_NOMOVE ) )
	{
		if (snapDistance)
		{
			if ( abs(lpwp->x) <= snapDistance )
				lpwp->x = 0;
			if ( abs(lpwp->y) <= snapDistance )
				lpwp->y = 0;
			if ( abs(lpwp->x + trayWidth - GetSystemMetrics(SM_CXSCREEN)) <= snapDistance )
				lpwp->x = GetSystemMetrics(SM_CXSCREEN) - trayWidth;
			if ( abs(lpwp->y + trayHeight - GetSystemMetrics(SM_CYSCREEN)) <= snapDistance )
				lpwp->y = GetSystemMetrics(SM_CYSCREEN) - trayHeight;
		}
		trayX = lpwp->x;
		trayY = lpwp->y;
	}

	if ( !( lpwp->flags & SWP_NOSIZE) )
	{
		setFirstLast();
		createBackground();

		InvalidateRect(hWnd, NULL, TRUE);
	}
	message.lResult = DefWindowProc(hWnd, message.uMsg, message.wParam, message.lParam);
}

void systray::onPaint (Message& message)
{
	HDC hdcBuffer;
	HBITMAP hbmBuffer, hbmBufOld;
	HRGN hrgn = NULL;

	PAINTSTRUCT ps;
	HDC	hdcScreen = BeginPaint(hWnd, &ps);

	hdcBuffer = CreateCompatibleDC(hdcScreen);
	hbmBuffer = CreateCompatibleBitmap(hdcScreen, trayWidth, trayHeight);
	hbmBufOld = (HBITMAP)SelectObject(hdcBuffer, hbmBuffer);

	if (transpBack)
		hrgn = CreateRectRgn( 0, 0, 0, 0 );

	// blit the skin
	paintBackground(hdcBuffer, hrgn);

	// blit the icons
	iconListPaint(hdcBuffer, hrgn);

	// set the windowRgn
	if (transpBack)
		SetWindowRgn(hWnd, hrgn, TRUE);

	// when the region is applied you can use a simple SRCCOPY from the buffer
	BitBlt(hdcScreen, 0, 0, trayWidth, trayHeight, hdcBuffer, 0, 0, SRCCOPY);

	// reselect the oldBitmap and kill the new one
	SelectObject(hdcBuffer, hbmBufOld);
	DeleteObject(hbmBuffer);

	// _tremove the memory DC
	DeleteDC(hdcBuffer);

	EndPaint(hWnd, &ps);
}

// bang commands!!
void systray::show(void)
{
	visible = TRUE;
	ShowWindow( hWnd, SW_SHOWNOACTIVATE );
}
void systray::hide(void)
{
	visible = FALSE;
	ShowWindow( hWnd, SW_HIDE );
}
void systray::toggle(void)
{
	visible = !IsWindowVisible( hWnd );
	ShowWindow( hWnd, visible ? SW_SHOWNOACTIVATE : SW_HIDE );
}
void systray::toggleOnTop(void)
{
	onTop = !(GetWindowLong(hWnd, GWL_EXSTYLE) & WS_EX_TOPMOST);
	SetWindowPos( hWnd, onTop ? HWND_TOPMOST : HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
}
void systray::move(int x, int y)
{
	SetWindowPos( hWnd, NULL, x, y, 0, 0, SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE );
}
void systray::size(int cx, int cy)
{
	SetWindowPos( hWnd, NULL, 0, 0, cx, cy, SWP_NOMOVE | SWP_NOZORDER | SWP_NOACTIVATE );
}

void systray::setFirstLast()
{
	// the following settings are used with the icon rectangles
	// to ensure they match the desired pattern
	firstX = ( ( direction == TD_LEFT ) || ( wrapDirection == TD_LEFT ) ) ?
	         ( trayWidth - borderRight - iconSpacingX - iconSize ) :
	         ( borderLeft + iconSpacingX );
	lastX = firstX + ( wrapCount - 1 ) * deltaX;

	firstY = ( ( direction == TD_DOWN ) || ( wrapDirection == TD_DOWN ) ) ?
	         ( borderTop + iconSpacingY ) :
	         ( trayHeight - borderBottom - iconSpacingY - iconSize );

	lastY = firstY + ( wrapCount - 1 ) * deltaY;
}

void systray::createBackground()
{
	HDC hdcScreen, hdcDst;
	HBITMAP hbmDstOld;

	// delete old objects if necessary...
	if ( hbmBack )
		DeleteObject( hbmBack );
	hbmBack = NULL;

	if ( hrgnBack )
		DeleteObject( hrgnBack );
	hrgnBack = NULL;

	// initiate a new destination dc
	hdcScreen = GetDC(hWnd);

	hdcDst = CreateCompatibleDC(hdcScreen);
	hbmBack = CreateCompatibleBitmap(hdcScreen, trayWidth, trayHeight);
	hbmDstOld = (HBITMAP)SelectObject(hdcDst, hbmBack);

	ReleaseDC(hWnd, hdcScreen);

	// if the background colors are needed put them on...
	if ((!hbmSkin) || transpSkin)
	{
		RECT r;
		HBRUSH brush, brushDk;

		// fill the border with the border colors
		if (borderTop || borderLeft || borderBottom || borderTop)
		{
			brush = CreateSolidBrush(clrBorder);
			brushDk = CreateSolidBrush(clrBorderDk);

			if (borderTop)
			{
				SetRect(&r, 0, 0, trayWidth, borderTop);
				FillRect(hdcDst, &r, brush);
			}
			if (borderLeft)
			{
				SetRect(&r, 0, borderTop, borderLeft, trayHeight - borderBottom);
				FillRect(hdcDst, &r, brush);
			}
			if (borderRight)
			{
				SetRect(&r, trayWidth - borderRight, borderTop, trayWidth, trayHeight - borderBottom);
				FillRect(hdcDst, &r, brushDk);
			}
			if (borderBottom)
			{
				SetRect(&r, 0, trayHeight - borderBottom, trayWidth, trayHeight);
				FillRect(hdcDst, &r, brushDk);
			}
			DeleteObject(brush);
		}

		// now fill the inner rectangle (excluding the borders) with the background color
		brush = CreateSolidBrush(clrBack);

		SetRect(&r, borderLeft, borderTop, trayWidth - borderRight, trayHeight - borderBottom);
		FillRect(hdcDst, &r, brush);

		DeleteObject(brush);
		DeleteObject(brushDk);
	}

	// if a background image is provided, put it on...
	if (hbmSkin)
	{

		HDC hdcSrc;
		HBITMAP hbmSrcOld;
		BITMAP bm;
		int innerWidth, innerHeight, bmInnerWidth, bmInnerHeight;

		// get info from the bitmap
		GetObject(hbmSkin, sizeof(BITMAP), &bm);

		// this is to prevent negative values on the width
		innerWidth = max(trayWidth - (borderLeft + borderRight), 0);
		innerHeight = max(trayHeight - (borderTop + borderBottom), 0);
		bmInnerWidth = max(bm.bmWidth - (borderLeft + borderRight), 1);
		bmInnerHeight = max(bm.bmHeight - (borderTop + borderBottom), 1);

		hdcSrc = CreateCompatibleDC(hdcDst);
		hbmSrcOld = (HBITMAP)SelectObject(hdcSrc, hbmSkin);

		if (borderTop || borderLeft || borderBottom || borderTop)
		{

			// paint the four corners first, and then the borders
			// topleft, topright, bottomright, bottomleft
			copyBlt(hdcDst, 0, 0, borderLeft, borderTop, hdcSrc, 0, 0);
			copyBlt(hdcDst, trayWidth - borderRight, 0, borderRight, borderTop, hdcSrc, bm.bmWidth - borderRight, 0);
			copyBlt(hdcDst, trayWidth - borderRight, trayHeight - borderBottom, borderRight, borderBottom, hdcSrc, bm.bmWidth - borderRight, bm.bmHeight - borderBottom);
			copyBlt(hdcDst, 0, trayHeight - borderBottom, borderLeft, borderBottom, hdcSrc, 0, bm.bmHeight - borderBottom);

			if (borderTop && innerWidth)
				sizeBlt(hdcDst, borderLeft, 0, innerWidth, borderTop, hdcSrc, borderLeft, 0, bmInnerWidth, borderTop);
			if (borderLeft && innerHeight)
				sizeBlt(hdcDst, 0, borderTop, borderLeft, innerHeight, hdcSrc, 0, borderTop, borderLeft, bmInnerHeight);
			if (borderRight && innerHeight)
				sizeBlt(hdcDst, trayWidth - borderRight, borderTop, borderRight, innerHeight, hdcSrc, bm.bmWidth - borderRight, borderTop, borderRight, bmInnerHeight);
			if (borderBottom && innerWidth)
				sizeBlt(hdcDst, borderLeft, trayHeight - borderBottom, innerWidth, borderBottom, hdcSrc, borderLeft, bm.bmHeight - borderBottom, bmInnerWidth, borderBottom);
		}

		// draw area in the middle
		if (innerWidth && innerHeight)
			sizeBlt(hdcDst, borderLeft, borderTop, innerWidth, innerHeight, hdcSrc, borderLeft, borderTop, bmInnerWidth, bmInnerHeight);

		SelectObject( hdcSrc, hbmSrcOld );
		DeleteDC( hdcSrc );
	}

	SelectObject( hdcDst, hbmDstOld );
	DeleteDC( hdcDst );

	if (!transpSkin)
	{
		transpBack = FALSE;
	}
	else
	{
		HRGN hOpaqueRgn = CreateRectRgn(0, 0, trayWidth, trayHeight);
		hrgnBack = BitmapToRegion(hbmBack, RGB(255, 0, 255), 0x101010, 0, 0);

		// to see if any pixels are transparent
		transpBack = !EqualRgn(hrgnBack, hOpaqueRgn);

		DeleteObject(hOpaqueRgn);

		if (!transpBack)
		{
			DeleteObject(hrgnBack);
			hrgnBack = NULL;
		}
	}
}

void systray::paintBackground( HDC hdcDst, HRGN hrgnDst )
{

	HDC hdcSrc;
	HBITMAP hbmOld;

	// if the background isn't there, make one...
	if (!hbmBack)
	{
		createBackground();
	}
	else
	{
		BITMAP bm;
		GetObject(hbmBack, sizeof(BITMAP), &bm);
		if (bm.bmWidth != trayWidth || bm.bmHeight != trayHeight)
		{
			createBackground();
		}
	}

	// if there is no background now... do nothing
	if (!hbmBack)
		return ;


	hdcSrc = CreateCompatibleDC(hdcDst);
	hbmOld = (HBITMAP)SelectObject(hdcSrc, hbmBack);

	BitBlt(hdcDst, 0, 0, trayWidth, trayHeight, hdcSrc, 0, 0, SRCCOPY);

	SelectObject(hdcSrc, hbmOld);
	DeleteDC(hdcSrc);

	if ( hrgnBack && transpBack )
		CombineRgn(hrgnDst, hrgnBack, NULL, RGN_COPY);

}

// grd Blitting functions...
// to enable transp / not transp painting
// and to enable tile / stretch resizing
int systray::copyBlt (HDC hdcDst, int xDst, int yDst, int cxDst, int cyDst, HDC hdcSrc, int xSrc, int ySrc)
{
	if (transpSkin)
		TransparentBltLS(hdcDst, xDst, yDst, cxDst, cyDst, hdcSrc, xSrc, ySrc, RGB(255, 0, 255));
	else
		return BitBlt(hdcDst, xDst, yDst, cxDst, cyDst, hdcSrc, xSrc, ySrc, SRCCOPY);

	return 0;
}

int systray::sizeBlt (HDC hdcDst, int xDst, int yDst, int cxDst, int cyDst, HDC hdcSrc, int xSrc, int ySrc, int cxSrc, int cySrc)
{
	HDC	hdcTmp;
	HBITMAP hbmTmp, hbmOld;
	int tmpX, tmpY;

	if (transpSkin)
	{
		hdcTmp = CreateCompatibleDC( hdcDst );
		hbmTmp = CreateCompatibleBitmap( hdcDst, cxDst, cyDst );
		hbmOld = (HBITMAP)SelectObject( hdcTmp, hbmTmp );
		tmpX = 0, tmpY = 0;
	}
	else
	{
		hdcTmp = hdcDst;
		tmpX = xDst, tmpY = yDst;
	}

	if (skinTiled)
	{
		int x, y;
		// fill the first row of images
		for (x = 0; x < cxDst; x += cxSrc)
			BitBlt(hdcTmp, tmpX + x, tmpY, min(cxSrc, cxDst - x), min(cySrc, cyDst), hdcSrc, xSrc, ySrc, SRCCOPY);
		// copy this row down the y axis...
		for (y = cySrc; y < cyDst; y += cySrc)
			BitBlt(hdcTmp, tmpX, tmpY + y, cxDst, min(cySrc, cyDst - y), hdcTmp, tmpX, tmpY, SRCCOPY);

	}
	else
	{
		// stretch it
		StretchBlt(hdcTmp, tmpX, tmpY, cxDst, cyDst, hdcSrc, xSrc, ySrc, cxSrc, cySrc, SRCCOPY);
	}

	// blit the resulting image onto the destination dc
	if (transpSkin)
	{
		TransparentBltLS(hdcDst, xDst, yDst, cxDst, cyDst, hdcTmp, tmpX, tmpY, RGB(255, 0, 255));

		// cleanup
		SelectObject( hdcTmp, hbmOld );
		DeleteObject( hbmTmp );
		DeleteDC( hdcTmp );
	}
	return 0;
}

PSYSTRAYICON systray::iconListFind(HWND hWnd, UINT uID)
{
	PSYSTRAYICON pSysTrayIcon = pFirst;

	while (pSysTrayIcon)
	{
		if ((pSysTrayIcon->hWnd == hWnd) && (pSysTrayIcon->uID == uID))
			return pSysTrayIcon;
		pSysTrayIcon = pSysTrayIcon->pNext;
	}
	return NULL;
}

PSYSTRAYICON systray::iconListFindPt(POINT pt)
{
	PSYSTRAYICON pSysTrayIcon = pFirst;

	while (pSysTrayIcon)
	{
		if ( PtInRect(&pSysTrayIcon->rc, pt) )
		{
			if ( IsWindow(pSysTrayIcon->hWnd) )
			{
				return pSysTrayIcon;
			}
			else
			{
				iconDelGRD(pSysTrayIcon);
				return NULL;
			}
		}
		pSysTrayIcon = pSysTrayIcon->pNext;
	}
	return NULL;
}

void systray::iconListCleanup()
{
	PSYSTRAYICON pSysTrayIcon = pFirst, pSysTrayIcon2 = NULL;

	while (pSysTrayIcon)
	{
		pSysTrayIcon2 = pSysTrayIcon->pNext;
		//char buffer2[MAX_LINE_LENGTH];
		//sprintf(buffer2,")checking window: %d"), (int)pSysTrayIcon->hWnd);
		//OutputDebugString(buffer2);

		if (!IsWindow(pSysTrayIcon->hWnd))
			iconDelGRD(pSysTrayIcon);

		pSysTrayIcon = pSysTrayIcon2;
	}
}

BOOL systray::iconListDelAll()
{

	PSYSTRAYICON pSysTrayIcon = pFirst, pSysTrayIcon2 = NULL;

	OnNumExe = 0;

	while (pSysTrayIcon)
	{
		// save the one to be deleted and move to the next
		pSysTrayIcon2 = pSysTrayIcon->pNext;
		// delete this history item
		iconDelGRD(pSysTrayIcon);
		pSysTrayIcon = pSysTrayIcon2;
	}
	return TRUE;
}

void systray::iconListPaint( HDC hdcDst, HRGN hrgnDst )
{
	PSYSTRAYICON pSysTrayIcon = pFirst;

	while (pSysTrayIcon)
	{
		if (pSysTrayIcon->hIcon)
		{
			DrawIconEx(hdcDst, pSysTrayIcon->rc.left, pSysTrayIcon->rc.top, pSysTrayIcon->hIcon, iconSize, iconSize, 0, NULL, DI_NORMAL);

			// combine the region
			if (pSysTrayIcon->hRgn && transpBack)
			{
				OffsetRgn(pSysTrayIcon->hRgn, pSysTrayIcon->rc.left, pSysTrayIcon->rc.top);
				CombineRgn(hrgnDst, hrgnDst, pSysTrayIcon->hRgn, RGN_OR);
				OffsetRgn(pSysTrayIcon->hRgn, -pSysTrayIcon->rc.left, -pSysTrayIcon->rc.top);
			}
		}
		pSysTrayIcon = pSysTrayIcon->pNext;
	}
}

/**** PROPERTY CHANGERS ****/

void systray::setIcon(PSYSTRAYICON pSysTrayIcon, HICON icon)
{
	if (pSysTrayIcon->hRgn)
		DeleteObject(pSysTrayIcon->hRgn);
	pSysTrayIcon->hRgn = NULL;

	if (pSysTrayIcon->hIcon)
		DestroyIcon(pSysTrayIcon->hIcon);
	pSysTrayIcon->hIcon = CopyIcon(icon);

	if (pSysTrayIcon->hIcon && transpBack)
	{
		pSysTrayIcon->hRgn = ::CreateRectRgn(0, 0, 0, 0);

		HDC hdc = ::CreateCompatibleDC( NULL );
		HBITMAP hbmNew = ::CreateCompatibleBitmap( hdc, iconSize, iconSize );
		HBITMAP hbmOld = (HBITMAP) ::SelectObject( hdc, hbmNew );

		if ( DrawIconEx(hdc, 0, 0, pSysTrayIcon->hIcon, iconSize, iconSize, 0, NULL, DI_MASK) )
		{
			int y = 0;
			while ( y < iconSize )
			{
				int x = 0;
				while ( x < iconSize )
				{
					UINT firstNTX = 0;
					// loop through all transparent pixels...
					while ( x < iconSize )
					{
						// if the pixel is white, then break
						if ( !::GetPixel(hdc, x, y) )
							break;
						x++;
					}
					// set first non transparent pixel
					firstNTX = x;
					// loop through all non transparent pixels
					while ( x < iconSize )
					{
						// if the pixel is white, then break
						if ( GetPixel(hdc, x, y) )
							break;
						x++;
					}
					// if found one or more non-transparent pixels in a row, add them to the rgn...
					if ((x - firstNTX) > 0)
					{
						HRGN hTempRgn = CreateRectRgn(firstNTX, y, x, y + 1);
						CombineRgn(pSysTrayIcon->hRgn, pSysTrayIcon->hRgn, hTempRgn, RGN_OR);
						DeleteObject(hTempRgn);
					}
					x++;

				}
				y++;

			}

		}
		SelectObject( hdc, hbmOld );
		DeleteObject( hbmNew );
		DeleteDC(hdc);
	}
}

void systray::setToolTip(PSYSTRAYICON pSysTrayIcon, char *szNewToolTip)
{
	BOOL existed = TRUE;
	UINT uStrLen = 0;
	TOOLINFO ti;

	ti.cbSize = sizeof(TOOLINFO);
	ti.hwnd	= hWnd;
	ti.hinst = NULL;

	if (!pSysTrayIcon->szTip)
	{
		existed = FALSE;
		pSysTrayIcon->uToolTip = uLastID++;
	}

	ti.uId = pSysTrayIcon->uToolTip;

	// delete old tip
	if (existed)
		delete[] pSysTrayIcon->szTip;

	// set the new tip
	if (szNewToolTip)
	{
		pSysTrayIcon->szTip = new char[strlen(szNewToolTip) + 1];
		strcpy(pSysTrayIcon->szTip, szNewToolTip);
	}
	else
	{
		pSysTrayIcon->szTip = NULL;
	}

	ti.lpszText = pSysTrayIcon->szTip;

	if ( existed )
	{
		SendMessage(tooltip, (pSysTrayIcon->szTip) ? TTM_UPDATETIPTEXT : TTM_DELTOOL, 0, (LPARAM)&ti);
	}
	else if ( pSysTrayIcon->szTip )
	{
		ti.uFlags = 0;
		ti.rect = pSysTrayIcon->rc;
		SendMessage(tooltip, TTM_ADDTOOL, 0, (LPARAM)&ti);
	}
}

void systray::adjustRect(PSYSTRAYICON pSysTrayIcon)
{
	BOOL existed = TRUE;
	PSYSTRAYICON pPrevTrayIcon = pSysTrayIcon->pPrev;

	if (pSysTrayIcon->pPrev)
	{
		if (direction & TD_VERTICAL)
		{
			// if there is room for one more icon
			if (pSysTrayIcon->pPrev->rc.top != lastY)
			{
				pSysTrayIcon->rc.top = pSysTrayIcon->pPrev->rc.top + deltaY;
				pSysTrayIcon->rc.left = pSysTrayIcon->pPrev->rc.left;
			}
			else
			{
				pSysTrayIcon->rc.top = firstY;
				pSysTrayIcon->rc.left = pSysTrayIcon->pPrev->rc.left + deltaX;
			}
		}
		else
		{
			// if there is room for one more icon
			if (pSysTrayIcon->pPrev->rc.left != lastX)
			{
				pSysTrayIcon->rc.left = pSysTrayIcon->pPrev->rc.left + deltaX;
				pSysTrayIcon->rc.top = pSysTrayIcon->pPrev->rc.top;
			}
			else
			{
				pSysTrayIcon->rc.left = firstX;
				pSysTrayIcon->rc.top = pSysTrayIcon->pPrev->rc.top + deltaY;
			}
		}
	}
	else
	{
		pSysTrayIcon->rc.left = firstX;
		pSysTrayIcon->rc.top = firstY;
	}

	pSysTrayIcon->rc.right = pSysTrayIcon->rc.left + iconSize;
	pSysTrayIcon->rc.bottom = pSysTrayIcon->rc.top + iconSize;

	if (pSysTrayIcon->szTip)
	{
		TOOLINFO ti;
		ti.hwnd = hWnd;
		ti.uId = pSysTrayIcon->uToolTip;
		ti.rect = pSysTrayIcon->rc;
		SendMessage(tooltip, TTM_NEWTOOLRECT, 0, (LPARAM)&ti);
	}
}

void systray::adjustSize()
{
	RECT rcOrg;
	int x, y, cx, cy;

	// this is the new size...
	if (pLast)
	{

		cx = iconSize + 2 * iconSpacingX + borderRight + borderLeft;
		cy = iconSize + 2 * iconSpacingY + borderBottom + borderTop;

		if (direction & TD_VERTICAL)
		{
			if (pLast->rc.left != firstX)
				cy += (direction == TD_DOWN) ? (lastY - firstY) : (firstY - lastY);
			else
				cy += (direction == TD_DOWN) ? (pLast->rc.top - firstY) : (firstY - pLast->rc.top);
			cx += (wrapDirection == TD_RIGHT) ? (pLast->rc.left - firstX) : (firstX - pLast->rc.left);
		}
		else
		{
			if (pLast->rc.top != firstY)
				cx += (direction == TD_RIGHT) ? (lastX - firstX) : (firstX - lastX);
			else
				cx += (direction == TD_RIGHT) ? (pLast->rc.left - firstX) : (firstX - pLast->rc.left);
			cy += (wrapDirection == TD_DOWN) ? (pLast->rc.top - firstY) : (firstY - pLast->rc.top);
		}
	}
	else
	{
		cx = borderLeft + borderRight;
		cy = borderTop + borderBottom;
	}

	// position it according to the old position
	GetWindowRect(hWnd, &rcOrg);

	x = ((direction == TD_RIGHT) || (wrapDirection == TD_RIGHT)) ? rcOrg.left : rcOrg.right - cx;
	y = ((direction == TD_DOWN) || (wrapDirection == TD_DOWN)) ? rcOrg.top : rcOrg.bottom - cy;

	SetWindowPos(hWnd, NULL, x, y, cx, cy, SWP_NOACTIVATE | SWP_NOZORDER);

	/*trayX = x;
	trayY = y;
	trayWidth = cx;
	trayHeight = cy;

	setFirstLast();
	createBackground();*/

	if (pLast)
	{
		PSYSTRAYICON pSysTrayIcon = pFirst;

		while (pSysTrayIcon)
		{
			adjustRect(pSysTrayIcon);
			pSysTrayIcon = (PSYSTRAYICON)pSysTrayIcon->pNext;
		}
	}
}

PSYSTRAYICON systray::iconAdd(PLSNOTIFYICONDATA pnid)
{
	PSYSTRAYICON pSysTrayIcon;

	if ( !IsWindow(pnid->hWnd) )
		return NULL;

	// check if the icon already exists
	pSysTrayIcon = iconListFind(pnid->hWnd, pnid->uID);

	// if found, modify instead of add
	if ( pSysTrayIcon )
		return iconMod( pnid );

	//PrintLog("iconAdd %s - (0x%.8X) - (0x%.8X)", pnid->szTip, pnid->hWnd, pnid->uID);

	// if not found (which is the way it should go)
	pSysTrayIcon = new SYSTRAYICON;
	memset(pSysTrayIcon, 0, sizeof(SYSTRAYICON));

	// transfer properties
	pSysTrayIcon->hWnd = pnid->hWnd;
	pSysTrayIcon->uID = pnid->uID;
	pSysTrayIcon->uCallbackMessage = ((pnid->uFlags & NIF_MESSAGE) ? pnid->uCallbackMessage : NULL);

	// set icon and region properties
	setIcon( pSysTrayIcon, (pnid->uFlags & NIF_ICON) ? pnid->hIcon : NULL );

	// add to last position in the list
	if (!pFirst)
	{
		pFirst = pSysTrayIcon;
		pSysTrayIcon->pPrev = NULL;
	}
	if (pLast)
	{
		pSysTrayIcon->pPrev = pLast;
		pLast->pNext = pSysTrayIcon;
	}
	pLast = pSysTrayIcon;
	pSysTrayIcon->pNext = NULL;

	// adjust the rectangle to the correct position ( uses list position )
	adjustRect(pSysTrayIcon);

	// add tooltip (uses rect)
	setToolTip(pSysTrayIcon, (pnid->uFlags & NIF_TIP) ? pnid->szTip : NULL);

	// remove "unwanted" icons
	iconListCleanup();

	if (autoSize)
		adjustSize(); // which will repaint the window
	else
		InvalidateRect(hWnd, &pSysTrayIcon->rc, TRUE);

	// if it's the first icon, send an SHOW message
	if ( (pFirst == pSysTrayIcon) && (hideIfEmpty) )
		ShowWindow(hWnd, SW_SHOWNOACTIVATE);

	// return the newly created icon

	OnNumCount++;

	if ( OnNumCount <= OnNumList.size() && OnNumExe ) {
		LSExecute(liteStep, OnNumList[(OnNumCount - 1)], NULL);
	}

	if ( OnAddCommand != NULL ) {
		LSExecute(liteStep, OnAddCommand, NULL);
	}

	return pSysTrayIcon;
}

PSYSTRAYICON systray::iconMod(PLSNOTIFYICONDATA pnid)
{
	PSYSTRAYICON pSysTrayIcon;

	// check if the icon already exists
	pSysTrayIcon = iconListFind(pnid->hWnd, pnid->uID);

	// if not found, add instead of modify
	if (!pSysTrayIcon)
		return iconAdd(pnid);

	pSysTrayIcon->uID = pnid->uID;

	if (pnid->uFlags & NIF_MESSAGE)
		pSysTrayIcon->uCallbackMessage = pnid->uCallbackMessage;

	if (pnid->uFlags & NIF_TIP)
		setToolTip(pSysTrayIcon, pnid->szTip);

	if (pnid->uFlags & NIF_ICON)
		setIcon(pSysTrayIcon, pnid->hIcon);

	// remove "unwanted" icons
	iconListCleanup();

	InvalidateRect(hWnd, NULL, TRUE);

	// return the "refreshed" icon
	return pSysTrayIcon;
}

BOOL systray::iconDel(PLSNOTIFYICONDATA pnid)
{
	PSYSTRAYICON pSysTrayIcon = iconListFind(pnid->hWnd, pnid->uID);

	if (pSysTrayIcon)
	{
		return iconDelGRD(pSysTrayIcon);
	}

	OnNumCount--;

	if ( OnNumCount <= OnNumList.size() && OnNumExe ) {
		LSExecute(liteStep, OnNumList[(OnNumCount - 1)], NULL);
	}

	if ( OnDelCommand != NULL ) {
		LSExecute(liteStep, OnDelCommand, NULL);
	}

	return FALSE;
}

BOOL systray::iconDelGRD(PSYSTRAYICON pSysTrayIcon)
{
	PSYSTRAYICON pIteratorSysTrayIcon;
	TOOLINFO ti;

	if ( !pSysTrayIcon )
		return FALSE;

	//PrintLog("iconDel %s - (0x%.8X) - (0x%.8X)", pSysTrayIcon->szTip, pSysTrayIcon->hWnd, pSysTrayIcon->uID);

	// remove the tooltip
	setToolTip(pSysTrayIcon, NULL);

	// remove the icon
	setIcon(pSysTrayIcon, NULL);

	// change all the rectangles and update the tooltips
	pIteratorSysTrayIcon = pLast;
	// swap rectangles... this is smooth
	while (pIteratorSysTrayIcon != pSysTrayIcon)
	{
		if (pIteratorSysTrayIcon && pIteratorSysTrayIcon->pPrev)
			pIteratorSysTrayIcon->rc = pIteratorSysTrayIcon->pPrev->rc;

		if (pIteratorSysTrayIcon->szTip)
		{
			ti.hwnd = hWnd;
			ti.uId = pIteratorSysTrayIcon->uToolTip;
			ti.rect = pIteratorSysTrayIcon->rc;
			SendMessage(tooltip, TTM_NEWTOOLRECT, 0, (LPARAM)&ti);
		}
		pIteratorSysTrayIcon = pIteratorSysTrayIcon->pPrev;
	}

	// if it's first or last, move those pointers
	if (pFirst == pSysTrayIcon)
		pFirst = pSysTrayIcon->pNext;
	else
		pSysTrayIcon->pPrev->pNext = pSysTrayIcon->pNext;

	if (pLast == pSysTrayIcon)
		pLast = pSysTrayIcon->pPrev;
	else
		pSysTrayIcon->pNext->pPrev = pSysTrayIcon->pPrev;

	// remove the icon class
	delete pSysTrayIcon;

	if (!pFirst && hideIfEmpty)
	{
		ShowWindow(hWnd, SW_HIDE);
	}
	else
	{
		if (autoSize)
			adjustSize();
		else
			InvalidateRect( hWnd, NULL, TRUE );
	}

	OnNumCount--;

	if ( OnNumCount <= OnNumList.size() && OnNumExe ) {
		LSExecute(liteStep, OnNumList[(OnNumCount - 1)], NULL);
	}

	if ( OnDelCommand != NULL ) {
		LSExecute(liteStep, OnDelCommand, NULL);
	}

	return TRUE;
}